%
%   INVPN1.M      Inverted Pendulum Demonstration #1 
%            (basic model simulations & classical control test)
%
%   This file simulates the inverted pendulum under several scenarios.  The
%   nonlinear and linear models are compared.  The Simulink linmod function is 
%   used to generate the linear model and this is compared to the analytically 
%   determined linear model.  A simple classical control scheme is attempted.  
%   The choice of a controller is done by trial and error.
%
%   This file is broken into three sections, as follows:
%     Part I.  Setup base data for the nonlinear & linear models and show that the
%              base plant is unstable.  Determine step response and compare the 
%              linear and nonlinear models.
%    Part II.  Use a Simulink S-function to represent the nonlinear equations. 
%              Use Simulink's SIM function to simulate nonlinear model and
%              Simulink's LINMOD command to extract the linear model around the 
%              desired reference point.
%   Part III.  Try to find a simple classical control scheme that stabilizes this 
%              system.  Nothing fancy here - just trial and error.  Then simulate 
%              the closed loop system behavior for a step change in the cart's 
%              position.
%
%   Related files:
%     invpnnl1.m   - contains basic nonlinear model for use in ODE45
%     invpnnl2.m   - contains nonlinear model written as S-function for Simulink
%     invpnsl.slx  - Simulink model containing link to S-function of nonlinear model
%     invpnsl1.slx - Simulink linear model with classical control of rod position
%     invpnsl2.slx - Simulink linear model with classical control of cart position
%     invpnsl3.slx - Simulink linear model with classical control of rod & cart position
%
%   Note: Additional M-files (INVPN2.M & INVPN3.M) look at this system with 
%         state feedback and a random disturbance as input.
%
%   File prepared by J. R. White, UMass-Lowell (last update: March 2020).
%

      clear all,   close all,   nfig = 0;
      format compact
      disp(' ')
      disp(' *** Summary Data from INVPN1.M ***')
      disp(' ')
%
%   Part I.  Setup base data for the nonlinear & linear models and show that the
%            base plant is unstable.  Determine step response and compare the 
%            linear and nonlinear models.
%
%   basic data
      M = 2.0;  m = 0.1;     % mass of cart and mass at end (Kg)
      len = 0.5;             % length of pendulum rod (m)
      g = 9.81;              % gravitational acceleration (m/s^2)
%
%  simulate a step change in input force u = 1 in nonlinear model
      u = 1;    to = 0;   tf = 1.0;
      zo = [0 0 0 0]';    tol = 1.0e-6;
      options = odeset('RelTol',tol);
      ftz = @(t,z) invpnnl1(t,z,u,M,m,g,len);
      [tnl1,znl1] = ode45(ftz,[to tf],zo,options);
%
%  create state space matrices for linear model (from analytical development)
      c1 = M*len;  c2 = m*len;  c3 = m*g;   c4 = (M+m)*g;
      A1 = [0  1  0  0;  c4/c1  0  0  0; 0  0  0  1; -c3/M  0  0  0];
      B1 = [0 -1/c1 0 1/M]';
      C1 = [1 0 0 0; 0  0  1  0];      D1 = [0 0]';
      disp('State Space Matrices for the Analytically Determined Linear Model')
      A1,  B1,  C1,  D1
%
%   compute eigenvalues of state matrix for analytical linear system
      disp('Eigenvalues of the "Analytical Linear Model"');     ev = eig(A1)
%
%   determine step response of the analytical linear system
      tl = linspace(to,tf,31); 
      sysl1 = ss(A1,B1,C1,D1);   [yl1,tl,zl1] = step(sysl1,tl);
%
%   plot results from nonlinear & linear models (no feedback)
      nfig = nfig+1;   figure(nfig);
      subplot(2,1,1),plot(tnl1,znl1(:,1),'g-',tl,zl1(:,1),'go','LineWidth',2),grid
      title('Inverted Pendulum Rod Angle (step response)')
      xlabel('Time (sec)'),ylabel('Rod Angle (radians)')
      legend('nonlinear','linear','Location','SouthWest')
%
      subplot(2,1,2),plot(tnl1,znl1(:,3),'r-',tl,zl1(:,3),'ro','LineWidth',2),grid
      title('Inverted Pendulum Cart Position (step response)')
      xlabel('Time (sec)'),ylabel('Cart Position (m)')
      legend('nonlinear','linear','Location','NorthWest')
%
%   Part II.  Use a Simulink S-function to represent the nonlinear equations. 
%             Use Simulink's SIM function to simulate nonlinear model and
%             Simulink's LINMOD command to extract the linear model around the 
%             desired reference point.
%
%  simulate a step change in input force u = 1 in nonlinear model
      invpnsl      % shows Simulink model with link to the nonlinear S-function model 
      ut = [to  u; tf  u];   % this sets the step input vector (SIM interpolates)
      options = simset('RelTol',tol);
      [tnl2,znl2,ynl2] = sim('invpnsl',[to tf],options,ut);
%
%  create state space matrices for linear model (from Simulink numerical development)
      [A2,B2,C2,D2] = linmod('invpnsl',zo,0);
      disp('State Space Matrices for the Numerically Determined Linear Model')
      A2,  B2,  C2,  D2
%
%   compute eigenvalues of state matrix for numerical linear system
      disp('Eigenvalues of the "Numerical Linear Model"');     ev = eig(A2)
%
%   determine step response of the numerical linear system
      tl = linspace(to,tf,31);   
      sysl2 = ss(A2,B2,C2,D2);   [yl2,tl,zl2] = step(sysl2,tl);
%
%   plot results from nonlinear & linear models from S-function (no feedback)
      nfig = nfig+1;   figure(nfig);
      subplot(2,1,1),plot(tnl2,znl2(:,1),'g-',tl,zl2(:,1),'go','LineWidth',2),grid
      title('S-Fun Inverted Pendulum Rod Angle (step response)')
      xlabel('Time (sec)'),ylabel('Rod Angle (radians)')
      legend('nonlinear','linear','Location','SouthWest')
%
      subplot(2,1,2),plot(tnl2,znl2(:,3),'r-',tl,zl2(:,3),'ro','LineWidth',2),grid
      title('S-Fun Inverted Pendulum Cart Position (step response)')
      xlabel('Time (sec)'),ylabel('Cart Position (m)')
      legend('nonlinear','linear','Location','NorthWest')
%
%   Part III.  Try to find a simple classical control scheme that stabilizes this 
%              system.  Nothing fancy here - just trial and error.  Then simulate 
%              the closed loop system behavior for a step change in the cart's 
%              position.
%
%   Note: MATLAB's root locus command is very specialized and can only treat 
%         SISO systems. 
%
%   Initial matrices for use in Simulink models
      A = A1;  B = B1;  C = C1;  D = D1;
%
%   Rod position feedback 
      invpnsl1      % shows simulink model (rod position feedback)
      disp('Following data for ROD POSITION feedback:');
      K1 = [-500 -100 -10 -1 -0.1 0.1 1 10 100 500];
      for j = 1:length(K1)
        k1 = K1(j),  [A1,B1,C1,D1] = linmod('invpnsl1');  eig(A1)
      end
%
%   Cart position feedback 
      invpnsl2      % shows simulink model (cart postion feedback)
      disp('Following data for CART POSITION feedback:');
      K2 = [-500 -100 -10 -1 -0.1 0.1 1 10 100 500];
      for j = 1:length(K2)
        k2 = K2(j),  [A2,B2,C2,D2] = linmod('invpnsl2');  eig(A2)
      end
%
%   now try adding two feedback loops (cart position outside and rod position inside)
      invpnsl3      % shows simulink model (both feedbacks)
      disp('Closed loop system with two feedback loops');
      k1 = -50,    k2 = -2      % default values of gains
      nfig = nfig+1;    cont = 'y';
      while cont == 'y'
        [A3,B3,C3,D3] = linmod('invpnsl3');      eig(A3)
%
%   step response of linear system with two feedback loops (cart goes from 0 -> 1)
        to = 0;   tf = 10;    tlf = linspace(to,tf,201);
        sysl3 = ss(A3,B3,C3,D3);   [ylf,tlf,zlf] = step(sysl3,tlf);
%
%   plot results from linear models (with two feedbacks)
        figure(nfig);
        subplot(3,1,1),plot(tlf,ylf(:,1),'g-','LineWidth',2),grid
        title(['Linear Inverted Pendulum Behavior (k1 = ', ...
           num2str(k1),' & k2 = ',num2str(k2),')'])
        ylabel('Rod Angle (radians)')
%
        subplot(3,1,2),plot(tlf,ylf(:,2),'r-','LineWidth',2),grid
        ylabel('Cart Position (m)')
%
        u = k2*(1-ylf(:,2)) - k1*ylf(:,1);
        subplot(3,1,3),plot(tlf,u,'b-','LineWidth',2),grid
        xlabel('Time (sec)'),ylabel('Input Force (N)')
%
        cont = input('Select different gains? (y/n): ', 's');
        if isempty(cont);   cont = 'n';    end
        if cont == 'y'
          disp('Input values for gains (k1 & k2)')
          k1 = input('k1 = ');     k2 = input('k2 = ');
        end
      end
%
%   now draw full size plots for last gains used
      nfig = nfig+1;    figure(nfig);
      plot(tlf,ylf(:,1),'g-','LineWidth',2),grid
      title(['Inverted Pendulum Rod Position (k1 = ', ...
              num2str(k1),' & k2 = ',num2str(k2),')'])
      xlabel('Time (sec)'),ylabel('Rod Angle (radians)')
%
      nfig = nfig+1;    figure(nfig);
      plot(tlf,ylf(:,2),'r-','LineWidth',2),grid
      title(['Inverted Pendulum Cart Position (k1 = ', ...
              num2str(k1),' & k2 = ',num2str(k2),')'])
      xlabel('Time (sec)'),ylabel('Cart Position (m)')
%
      nfig = nfig+1;    figure(nfig);
      plot(tlf,u,'b-','LineWidth',2),grid
      title(['Inverted Pendulum Input Force (k1 = ', ...
              num2str(k1),' & k2 = ',num2str(k2),')'])
      xlabel('Time (sec)'),ylabel('Input Force (N)')
%
      disp('End of simulation')
%
%   end of simulation
%