%
%   INVPN2.M    Inverted Pendulum Demonstration #2
%
%   This file simulates the inverted pendulum (linear and nonlinear model) with
%   state feedback control.  File invpnnl3.m is used for the nonlinear plant model.
%
%   This file is broken into five sections, as follows:
%     Part I. Setup base data for the linear model and show that the base
%             plant is unstable.
%    Part II. Add state feedback control to the linear model to stabilize the system. 
%             Simulate system behavior for a step change in the cart's position.
%   Part III. Add state feedback control and a full observer to stabilize the system.
%             Simulate system behavior for a step change in the cart's position.
%             This should give the same simulation as Part II.
%    Part IV. Same as Part III using nonlinear model for the plant and the
%             linear model for the state observer.
%     Part V. Compute and plot (ie. Bode plots) the transfer function for the
%             closed loop system.  Here we are interested in the dynamics of the
%             cart's position relative to a change in the set point (desired position).
%             Since the closed loop dynamics are identical, we will use the case
%             without the state estimator.
%
%   File prepared by J. R. White, UMass-Lowell (last update: March 2020).
%

      clear all,   close all,   nfig = 0;
      format compact
      disp(' ')
      disp(' *** Summary Data from INVPN2.M ')
      disp(' ')
%
%   Part I. Setup base data for the linear model and show that the base
%           plant is unstable.
%
%   basic data
      M = 2.0;  m = 0.1;     % mass of cart and mass at end (Kg)
      len = .5;              % length of pendulum rod (m)
      g = 9.81;              % gravitational acceleration (m/s^2)
%
%  create state space matrices for linear model (output cart position)
      c1 = M*len;  c2 = m*len;  c3 = m*g;   c4 = (M+m)*g;
      A = [0  1  0  0;  c4/c1  0  0  0; 0  0  0  1; -c3/M  0  0  0];
      B = [0 -1/c1 0 1/M]';
      C = [0  0  1  0];      D = [0];
      disp('State Space Matrices for the Linear Model')
      A,  B,  C,  D
%
%   compute eigenvalues of state matrix for linear system
      disp('Eigenvalues of the "Linear Model"');     ev = eig(A)
%
%   Part II. Add state feedback control to stabilize the system.  Simulate system 
%            behavior for a step change in the cart's position.
%
%   check for full state controllability
      disp('Controllability Matrix for this system'),   CM = ctrb(A,B)
      disp('Rank of Controllability Matrix'),   rank(CM)     
%
%   calculate state feedback gains for specified closed loop poles
      clp = [-1.5+3.0j  -1.5-3.0j  -5.0  -6.0];
      Ks = place(A,B,clp);
      disp('Desired closed loop poles for state feedback controller');   clp
      disp('State feedback gains needed to give desired poles');         Ks
      disp('Calculated eigenvalues of system with state feedback'); eig(A-B*Ks)
%
%   calculate Nr for zero SS error  (see derivation in notes)
      Nr = -1.0/(C*inv(A-B*Ks)*B);
      disp('Setpoint gain for zero SS error');   Nr
%
%   simulate linear plant + controller
      tto = 0;  ttf = 5;   t = linspace(tto,ttf,101);
      syscl1 = ss(A-B*Ks,B*Nr,C,D);   [y1,t,x1] = step(syscl1,t);
%
%   plot results from state feedback case
      nfig = nfig+1;      figure(nfig)
      plot(t,y1,'r-','LineWidth',2),grid
      xlabel('Time (sec)'),ylabel('Cart Position (m)')
      title('Inverted Pendulum with State Control (Cart Position)')
%
%   also plot state variables
      nfig = nfig+1;      figure(nfig)
      subplot(4,1,1),plot(t,x1(:,1),'g-','LineWidth',2),grid,ylabel('Angle')
      title('States for Inverted Pendulum (State Feedback)')
      subplot(4,1,2),plot(t,x1(:,2),'g-','LineWidth',2),grid,ylabel('d(Angle)/dt')
      subplot(4,1,3),plot(t,x1(:,3),'r-','LineWidth',2),grid,ylabel('Position')
      subplot(4,1,4),plot(t,x1(:,4),'r-','LineWidth',2),grid,ylabel('d(Pos)/dt')
      xlabel('Time (sec)')
%
%   Part III. Add state feedback control and a full observer to stabilize the system.
%             Simulate system behavior for a step change in the cart's position.  
%             This should give the same simulation as Part II.
%
%   check for full state observability
      disp('Observability Matrix for this system'),   OM = obsv(A,C)
      disp('Rank of Observability Matrix'),   rank(OM)     
%
%   calculate estimator gains for specified observer poles
      op = 2*clp;    % estimator dynamics is 2 times faster than closed loop poles
      L = place(A',C',op);   L = L';
      disp('Desired observer poles for state feedback controller');   op
      disp('Estimator gains needed to give desired poles');           L
      disp('Calculated eigenvalues of estimator system');      eig(A-L*C)

%   setup matrices for linear plant + controller model
      A11 = A;     A12 = -B*Ks;         B1 = B*Nr;
      A21 = L*C;   A22 = A-L*C-B*Ks;    B2 = B*Nr;
      AB = [A11 A12; A21 A22];   BB = [B1; B2];
      zz = 0;   CB = [C zz*C];
%
%   simulate linear plant + controller
      syscl2 = ss(AB,BB,CB,D);   [y2,t,x2] = step(syscl2,t);
%
%   separate plant and estimator states
      nn = max(size(A));
      xp2 = x2(:,1:nn);     xe2 = x2(:,nn+1:2*nn);
%
%   plot results from case with full observer
      nfig = nfig+1;      figure(nfig)
      plot(t,y2,'r-','LineWidth',2),grid
      xlabel('Time (sec)'),ylabel('Cart Position (m)')
      title(['Inverted Pendulum with State Control & Full Observer (Cart Position)'])
%
%   also plot state variables
      nfig = nfig+1;      figure(nfig)
      subplot(4,1,1),plot(t,xp2(:,1),'g-','LineWidth',2),grid,ylabel('Angle')
      title(['States for Inverted Pendulum (State Feedback/Full Observer)'])
      subplot(4,1,2),plot(t,xp2(:,2),'g-','LineWidth',2),grid,ylabel('d(Angle)/dt')
      subplot(4,1,3),plot(t,xp2(:,3),'r-','LineWidth',2),grid,ylabel('Position')
      subplot(4,1,4),plot(t,xp2(:,4),'r-','LineWidth',2),grid,ylabel('d(Pos)/dt')
      xlabel('Time (sec)')
%
%   also plot error for plant vs estimator
      nfig = nfig+1;      figure(nfig)
      subplot(4,1,1),plot(t,xp2(:,1)-xe2(:,1),'g-','LineWidth',2),grid,ylabel('Angle')
      title(['Difference Between Plant and Observer States'])
      subplot(4,1,2),plot(t,xp2(:,2)-xe2(:,2),'g-','LineWidth',2),grid,ylabel('d(Angle)/dt')
      subplot(4,1,3),plot(t,xp2(:,3)-xe2(:,3),'r-','LineWidth',2),grid,ylabel('Position')
      subplot(4,1,4),plot(t,xp2(:,4)-xe2(:,4),'r-','LineWidth',2),grid,ylabel('d(Pos)/dt')
      xlabel('Time (sec)')
%
%   Part IV. Same as Part III using nonlinear model for the plant and the
%            linear model for the state observer.
%
%   use state feedback & estimator gains from above linear simulations
%
%   simulate linear plant + controller
      rd = 1;   % unit step in cart position setpoint
      tt0 = 0;  ttf = 5;   Zo = zeros(8,1);   tol = 1.0e-6;
      options = odeset('RelTol',tol);
      ftz = @(t,Z) invpnnl3(t,Z,M,m,g,len,A,B,C,D,Ks,Nr,L,rd);
      [tt,Z] = ode45(ftz,[tto ttf],Zo,options);
%
%   separate plant and estimator states
      nn = max(size(A));
      zp =Z(:,1:nn);     ze = Z(:,nn+1:2*nn);
%
%   plot results from nonlinear case with full observer
      nfig = nfig+1;      figure(nfig)
      plot(tt,zp(:,3),'r-','LineWidth',2),grid
      xlabel('Time (sec)'),ylabel('Cart Position (m)')
      title('Nonlinear Inverted Pendulum with State Control & Full Observer (Cart Position)')
%
%   also plot state variables
      nfig = nfig+1;      figure(nfig)
      subplot(4,1,1),plot(tt,zp(:,1),'g-','LineWidth',2),grid,ylabel('Angle')
      title('States for Nonlinear Inverted Pendulum (State Feedback/Full Observer)')
      subplot(4,1,2),plot(tt,zp(:,2),'g-','LineWidth',2),grid,ylabel('d(Angle)/dt')
      subplot(4,1,3),plot(tt,zp(:,3),'r-','LineWidth',2),grid,ylabel('Position')
      subplot(4,1,4),plot(tt,zp(:,4),'r-','LineWidth',2),grid,ylabel('d(Pos)/dt')
      xlabel('Time (sec)')
%
%   also plot error for plant vs estimator
      nfig = nfig+1;      figure(nfig)
      subplot(4,1,1),plot(tt,zp(:,1)-ze(:,1),'g-','LineWidth',2),grid,ylabel('Angle')
      title(['Difference Between Nonlinear Plant and Observer States'])
      subplot(4,1,2),plot(tt,zp(:,2)-ze(:,2),'g-','LineWidth',2),grid,ylabel('d(Angle)/dt')
      subplot(4,1,3),plot(tt,zp(:,3)-ze(:,3),'r-','LineWidth',2),grid,ylabel('Position')
      subplot(4,1,4),plot(tt,zp(:,4)-ze(:,4),'r-','LineWidth',2),grid,ylabel('d(Pos)/dt')
      xlabel('Time (sec)')
%
%  Part V. Compute and plot (ie. Bode plots) the transfer function for the
%          closed loop system.  Here we are interested in the dynamics of the
%          cart's position relative to a change in the set point (desired position).
%          Since the closed loop dynamics are identical, we will use the case
%          without the state estimator.
%
%   evaluate freqency response
      freqp = input('Perform frequency domain analysis? (y/n) [n]:  ','s');
      if isempty(freqp);   freqp = 'n';    end
      if freqp == 'y'
%
%   create Bode plots
        nfig = nfig+1;   figure(nfig)
        w = logspace(-1,2,100);   bode(syscl1,w),  grid 
        hd = findobj(gcf,'Color','blue'); nhd = length(hd); 
        for i = 1:nhd; set(hd(i),'LineWidth',2); end
        title('G(s) = X(s)/R(s)  (Cart Position)')
%
%   special note
        disp('  ')
        disp('Note that this system is non minimum phase because of the zero')
        disp('in the right portion of the complex plane')
%
%   convert from state-space to transfer function form
        disp('  ')
        disp('G(s) for Cart Position (Transfer Function form)')
        [num1,den1] = tfdata(syscl1,'v')
        printsys(num1,den1)
%
%   convert from state-space to zero-pole-gain form
        disp('  ')
        disp('G(s) for Cart Position (Zero-Pole form)')
        [z1,p1,k1] = zpkdata(syscl1,'v')
        zpk(syscl1)
%
%   finally, compute the damping ratio and natural frequency for the
%   closed loop system and make a pole zero plot
        disp('  ')
        disp('Damping and natural frequencies for closed loop system'), damp(p1)
        nfig = nfig+1;   figure(nfig)
        pzmap(syscl1),sgrid
        hd = findobj(gcf,'Color','blue'); nhd = length(hd); 
        for i = 1:nhd; set(hd(i),'LineWidth',2); end
        title('Pole-Zero Map for Closed Loop System')
      end     %  generate above plots if freqp = 'y'
%
%   end of simulation