%
%   INVPN3.M    Inverted Pendulum Demonstration #3 (with disturbance input)
%
%   This file simulates the inverted pendulum (linear model) with state
%   feedback control.  It is similar to INVPN2.M except this simulation contains
%   a disturbance input.  We will use the state gains from the standard design 
%   with this case.  Let's see how this works...
%
%   File prepared by J. R. White, UMass-Lowell (last update: March 2020)
%

      clear all,   close all,   nfig = 0;
%
      disp(' ')
      disp(' *** Summary Data from INVPN3.M ***')
      disp(' ')
%
%   basic data
      M = 2.0;  m = 0.1;     % mass of cart and mass at end (kg)
      len = .5;              % length of pendulum rod (m)
      g = 9.81;              % gravitational acceleration (m/s^2)
%
%  create state space matrices for linear model (output cart position)
      c1 = M*len;  c2 = m*len;  c3 = m*g;   c4 = (M+m)*g;
      A = [0  1  0  0;  c4/c1  0  0  0; 0  0  0  1; -c3/M  0  0  0];
      B1 = [0 -1/c1 0 1/M]';   B2 = [0 1/c2 0 0]';
      C = [0  0  1  0];      D = [0];
      disp('State Space Matrices for the Linear Model')
      A,  B1,  B2,  C
%
%   check for full state controllability (SISO system)
     disp('Controllability Matrix for this system'),   CM = ctrb(A,B1)
     disp('Rank of Controllability Matrix'),   rank(CM)     
%
%   calculate state feedback gains for specified closed loop poles (SISO system)
      clp = [-1.5+3.0j  -1.5-3.0j  -5.0  -6.0];
      Ks = place(A,B1,clp);
      disp('Desired closed loop poles for state feedback controller');     clp
      disp('State feedback gains needed to give desired poles');           Ks
      disp('Calculated eigenvalues of system with state feedback'); eig(A-B1*Ks)
%
%   calculate Nr for zero SS error with no disturbance  (see derivation in notes)
      Nr = -1.0/(C*inv(A-B1*Ks)*B1);
      disp('Setpoint gain for zero SS error');    Nr
%
%   simulate linear plant + controller (use lsim for consistency in all cases)
      BB = [B1 B2];   D = [0 0];   % two inputs and one output
      syscl = ss(A-B1*Ks,BB,C,D);
%
%   Case 1: unit step change in cart position with no disturbance
%
      to = 0;  tf = 5;   Nt = 101;   t = linspace(to,tf,Nt)'; 
      u1 = zeros(size(t));      % controlled input (initialize)
      rd = ones(size(t));       % step change in setpoint
      v1 = zeros(size(t));      % disturbance input (zero for Case 1)
      w1 = [Nr*rd v1];
      [y1,t,x1] = lsim(syscl,w1,t); 
      for i = 1:Nt,  u1(i) = Nr*rd(i)-Ks*x1(i,:)';   end    % controlled input 
%
%   plot results from Case 1
      nfig = nfig+1;      figure(nfig)
      subplot(2,1,1),plot(t,y1,'r-','LineWidth',2),grid,ylabel('Cart Position (m)')
      title('Linear Inverted Pendulum (Case 1: rd = 1 & v = 0)')
      subplot(2,1,2),plot(t,u1,'g--',t,10*v1,'b-','LineWidth',2),grid
      ylabel('Inputs (N)'),xlabel('Time (sec)')
      legend('u(t)','10*v(t)','Location','SouthEast')
%
%   Case 2: no change in cart position with constant disturbance (0.2 N)
%
      u2 = zeros(size(t));          % controlled input (initialize)
      rd = zeros(size(t));          % no change in setpoint
      v2 = 0.2*ones(size(t));       % disturbance input (0.2N)
      w2 = [Nr*rd v2];
      [y2,t,x2] = lsim(syscl,w2,t); 
      for i = 1:Nt,  u2(i) = Nr*rd(i)-Ks*x2(i,:)';   end    % controlled input 
%
%   plot results from Case 2
      nfig = nfig+1;      figure(nfig)
      subplot(2,1,1),plot(t,y2,'r-','LineWidth',2),grid,ylabel('Cart Position (m)')
      title('Linear Inverted Pendulum (Case 2: rd = 0 & v = 0.2N)')
      subplot(2,1,2),plot(t,u2,'g--',t,10*v2,'b-','LineWidth',2),grid
      ylabel('Inputs (N)'),xlabel('Time (sec)')
      legend('u(t)','10*v(t)')
%
%   Case 3: unit step change in cart position with constant disturbance (0.2 N)
%
      u3 = zeros(size(t));          % controlled input (initialize)
      rd = ones(size(t));           % step change in setpoint
      v3 = 0.2*ones(size(t));       % disturbance input (0.2N)
      w3 = [Nr*rd v3];
      [y3,t,x3] = lsim(syscl,w3,t); 
      for i = 1:Nt,  u3(i) = Nr*rd(i)-Ks*x3(i,:)';   end    % controlled input 
%
%   plot results from Case 3
      nfig = nfig+1;      figure(nfig)
      subplot(2,1,1),plot(t,y3,'r-','LineWidth',2),grid,ylabel('Cart Position (m)')
      title('Linear Inverted Pendulum (Case 3: rd = 1 & v = 0.2N)')
      subplot(2,1,2),plot(t,u3,'g--',t,10*v3,'b-','LineWidth',2),grid
      ylabel('Inputs (N)'),xlabel('Time (sec)')
      legend('u(t)','10*v(t)')
%
%   Case 4: unit step change in cart position with random disturbance (+/- 0.2 N)
%
      u4 = zeros(size(t));        % controlled input (initialize)
      rd = ones(size(t));         % step change in setpoint
      rn = rand(size(t));   a = -0.2;   b = 0.2;      
      v4 = (b-a)*rn + a;          % should be uniformly distributed between +/- 0.2 N 
      w4 = [Nr*rd v4];
      [y4,t,x4] = lsim(syscl,w4,t); 
      for i = 1:Nt,  u4(i) = Nr*rd(i)-Ks*x4(i,:)';   end    % controlled input 
%
%   plot results from Case 4
      nfig = nfig+1;      figure(nfig)
      subplot(2,1,1),plot(t,y4,'r-','LineWidth',2),grid,ylabel('Cart Position (m)')
      title('Linear Inverted Pendulum (Case 4: rd = 1 & v = random noise (+/-0.2N))')
      subplot(2,1,2),plot(t,u4,'g--',t,10*v4,'b-','LineWidth',2),grid
      ylabel('Inputs (N)'),xlabel('Time (sec)')
      legend('u(t)','10*v(t)','Location','SouthEast')
%
%   end of simulation
